from dataclasses import dataclass
from .enums import SumTypes, TimePeriods
from ...services.locale.querytypes.queryservice import QueryService

@dataclass(frozen=True)
class QueryProperties:
    sumtype: SumTypes
    timecalc:TimePeriods
    cycle: TimePeriods
    queryservice: QueryService


from dataclasses import dataclass, field

@dataclass(frozen=False)
class ChargerEntitiesModel:
    powermeter: str
    powerswitch: str
    ampmeter: str = ""
    chargerentity: str = ""
    entityschema: str = ""
    imported_entities:list = field(default_factory=lambda: [])
    imported_entityendings:list = field(default_factory=lambda: [])
    


from dataclasses import dataclass, field
from typing import List, Dict
from .hourobject import HourObject

@dataclass(frozen=False)
class HoursModel:
    non_hours: List[int] = field(default_factory=lambda : [])
    caution_hours: List[int] = field(default_factory=lambda : [])
    dynamic_caution_hours: Dict[int, float] = field(default_factory=lambda : {})
    hours_today: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    hours_tomorrow: HourObject = field(default_factory=lambda : HourObject([],[],dict()))


@dataclass(frozen=False)
class HourSelectionOptions:
    cautionhour_type: float = 0
    absolute_top_price: float = 0
    min_price: float = 0
    allow_top_up: bool = False
    conserve_top_up: bool = False


@dataclass(frozen=False)
class HourSelectionModel:
    prices_today: List[float] = field(default_factory=lambda : [])
    prices_tomorrow: List[float] = field(default_factory=lambda : [])
    hours: HoursModel = HoursModel()
    options: HourSelectionOptions = HourSelectionOptions



from ...models.hourselection.hourobject import HourObject
from ...models.hourselection.topupdto import HoursDTO, TopUpDTO
import operator

def top_up(model:TopUpDTO) -> HourObject:
    if model.prices_tomorrow is None or len(model.prices_tomorrow) == 0:
        return

    today = list(model.prices[model.hour-1:23])
    tomorrow = list(model.prices_tomorrow[0:model.hour-1])
    cheap_max = 0

    if max(today) < (sum(tomorrow)/len(tomorrow)):
        is_today = True
        cheap_max = max(today)
    elif max(tomorrow) < (sum(today)/len(today)):
        is_today = False
        cheap_max = max(tomorrow)
    
    if cheap_max == 0 or len(today) < 9:
        return HourObject(
        nh=model.nh,
        ch=model.ch,
        dyn_ch=model.dyn_ch
        )
    
    result = _remove_and_add_for_top_up(
        model=HoursDTO(
            model.nh, 
            model.ch, 
            model.dyn_ch, 
            model.top_price, 
            model.min_price),
        today_dict=_create_partial_dict(
            today,
            model.hour, 
            True), 
        tomorrow_dict=_create_partial_dict(
            tomorrow, 
            model.hour, 
            False), 
        max_cheap=cheap_max, 
        today=is_today
        )

    result.nh = list(set(result.nh))
    result.nh.sort()
    try:
        dynamic_caution_hours = {key: value for (key, value) in result.dyn_ch.items() if key not in model.nh}
    except Exception as e:
        print(f"{e}: {result.dyn_ch}")
        dynamic_caution_hours = {}

    return HourObject(
        nh=result.nh,
        ch=[],
        dyn_ch=dynamic_caution_hours
    )

def _remove_and_add_for_top_up(
    model: HoursDTO,
    today_dict:dict, 
    tomorrow_dict:dict, 
    max_cheap:float, 
    today:bool
    ) -> HoursDTO:

    removed = 0
    remove_lists = dict(
        nh_remove= [],
        ch_remove= [],
        dyn_ch_remove= []
    )
        
    if today is True:
        removedict = today_dict
        adddict = tomorrow_dict
    else:
        removedict = tomorrow_dict
        adddict = today_dict

    removed += _append_items_to_remove(model, remove_lists, removedict)
    _remove_items(remove_lists, model)
    
    sorted_add = list(dict(sorted(adddict.items(), key=operator.itemgetter(1),reverse=True)).keys())
    added = 0
    for i in range(0, removed-1):
        try:
            if adddict[sorted_add[i]] > max_cheap:
                if sorted_add[i] not in model.nh or sorted_add[i] in model.dyn_ch.keys():
                    model.nh.append(sorted_add[i])
                    model.dyn_ch.pop(i)
                    added += 1
        except:
            continue
    if added == 0:
        for a in adddict:
            if a in model.dyn_ch.keys():
                model.nh.append(a)
                model.dyn_ch.pop(a)
    return model

def _append_items_to_remove(model:HoursDTO, remove_lists:dict, removedict:dict) -> int:
    ret = 0

    def _append_non_hours(model:HoursDTO, nonhours_remove:list, removedict:dict) -> int:
        _removed = 0
        for i in [i for i in model.nh if i in removedict.keys() and removedict[i] < model.top_price]:
            if i in removedict.keys():
                nonhours_remove.append(i)
                _removed += 1
        return _removed

    def _append_caution_hours(cautionhours:list, cautionhours_remove:list, removedict:dict) -> int:
        _removed = 0
        for i in cautionhours:
            if i in removedict.keys():
                cautionhours_remove.append(i)
                _removed += 1
        return _removed

    def _append_dyn_caution_hours(dyn_ch: dict, dyn_cautionhours_remove:list, removedict:dict) -> int:
        _removed = 0
        for k,v in dyn_ch.items():
            if k in removedict.keys():
                dyn_cautionhours_remove.append(k)
                _removed += 1
        return _removed

    ret += _append_non_hours(model, remove_lists["nh_remove"], removedict)
    ret += _append_caution_hours(model.ch, remove_lists["ch_remove"], removedict)
    ret += _append_dyn_caution_hours(model.dyn_ch, remove_lists["dyn_ch_remove"], removedict)
    return ret

def _remove_items(remove_lists:dict, model: HoursDTO) -> None:
    
    def _remove_from_list(checklist: list, deletelist: list|dict):
        if len(checklist) > 0:
            for i in checklist:
                if isinstance(deletelist, list):
                    deletelist.remove(i)
                elif isinstance(deletelist, dict):
                    deletelist.pop(i)    
    
    _remove_from_list(remove_lists["nh_remove"],model.nh)
    _remove_from_list(remove_lists["ch_remove"],model.ch)
    _remove_from_list(remove_lists["dyn_ch_remove"], model.dyn_ch)

def _create_partial_dict(input: list, hour:int, today:bool = True) -> dict:
    ret = dict()
    if today:
        dictrange = range(hour,24)
    else:
        dictrange = range(0,hour-1)
    assert len(dictrange) == len(input)
    
    for idx, val in enumerate(input):
        ret[dictrange[idx]] = val
    return ret


import logging
from .hubmember import HubMember

_LOGGER = logging.getLogger(__name__)


class ChargerSwitch(HubMember):
    def __init__(
            self,
            hass,
            data_type: type,
            listenerentity,
            initval,
            currentname: str,
            ampmeter_is_attribute: bool,
            hubdata=None,
            init_override:bool = False
    ):
        self._hubdata = hubdata
        self._hass = hass
        self._value = initval
        self._current = None
        self._current_attr_name = currentname
        self._ampmeter_is_attribute = ampmeter_is_attribute
        super().__init__(data_type=data_type, listenerentity=listenerentity, init_override=init_override, initval=initval)

    @property
    def is_initialized(self) -> bool:
        if self._hubdata.chargerobject is not None and not self._hubdata.chargerobject.is_initialized:
            return False
        return super().is_initialized

    @property
    def current(self) -> int:
        if self._current is None:
            return 0
        return self._current

    @current.setter
    def current(self, value):
        try:
            self._current = int(value)
        except ValueError:
            pass

    def updatecurrent(self):
        if self._hubdata.chargertype.charger.options.charger_is_outlet is True:
            pass
        elif self._ampmeter_is_attribute is True:
            ret = self._hass.states.get(self.entity)
            if ret is not None:
                ret_attr = str(ret.attributes.get(self._current_attr_name))
                self.current = ret_attr
            else:
                self._log_warning_once()
        else:
            ret = self._hass.states.get(self._current_attr_name)
            if ret is not None:
                self.current = ret.state
            else:
                self._log_warning_once()

    HASLOGGED_INITWARN = False
    def _log_warning_once(self):
        if not self.HASLOGGED_INITWARN:
            _LOGGER.warning("Chargerobject state was None while getting current. "
                "Updatecurrent will still process but you may see a missmatch in frontend.")
            self.HASLOGGED_INITWARN = True



import statistics as stat


class HourSelectionHelpers:
    @staticmethod
    def _create_dict(input: list):
        ret = dict()
        for idx, val in enumerate(input):
            ret[idx] = val
        try:
            assert len(ret) == 24
        except Exception:
            raise ValueError
        return ret

    @staticmethod
    def _try_parse(input:str, parsetype:type):
        try:
            ret = parsetype(input)
            return ret
        except:
            return False
    
    @staticmethod
    def _convert_none_list(lst: any) -> list:
        ret = []
        if lst is None or not isinstance(lst, list):
            return ret
        try:
            for l in lst:
                if l is None:
                    return ret
            return lst
        except:
            return HourSelectionHelpers._make_array_from_empty(lst)

    @staticmethod
    def _make_array_from_empty(input) -> list:
        array = input.split(",")
        list = [p for p in array if len(p) > 0]
        ret = []
        if len(list) > 24:
            try:
                for l in list:
                    parsed_item = HourSelectionHelpers._try_parse(l, float)
                    if not parsed_item:
                        parsed_item = HourSelectionHelpers._try_parse(l, int)
                    assert type(parsed_item) is float or type(parsed_item) is int
                    ret.append(parsed_item)
                return ret
            except:
                return []
        return []

    @staticmethod
    def _normalize_prices(prices:list) -> list:
        min_price = min(prices)
        c = 0
        if min_price <= 0:
            c = abs(min_price) + 0.01
        ret = []
        for p in prices:
            pp = p+c
            divider = min_price if min_price > 0 else c
            ret.append(pp/divider)
        return ret

    @staticmethod
    def _rank_prices(hourdict: dict, normalized_hourdict: dict) -> dict:
        ret = dict()
        _maxval = max(hourdict.values())
        _max_normalized = max(normalized_hourdict.values())
        peaqstdev = _maxval/abs(_max_normalized/stat.stdev(normalized_hourdict.values()))
        
        if peaqstdev < min(hourdict.values()):
            peaqstdev = peaqstdev + min(hourdict.values())
        for key in hourdict:
            if hourdict[key] > peaqstdev:
                _permax = round(hourdict[key] / _maxval, 2)
                ret[key] = {"val": hourdict[key], "permax": _permax}
        return HourSelectionHelpers._discard_excessive_hours(ret)

    @staticmethod
    def _discard_excessive_hours(hours: dict):
        """There should always be at least four regular hours before absolute_top_price kicks in."""
        if len(hours) < 20:
            return hours
        while len(hours) >= 20:
            to_pop = dict(sorted(hours.items(), key=lambda item: item[1]['val']))    
            hours.pop(list(to_pop.keys())[0])
        return hours


from dataclasses import dataclass
from .hourobject import HourObject

@dataclass
class HoursDTO:
    nh: list
    ch: list
    dyn_ch: dict
    top_price:float
    min_price:float

@dataclass
class TopUpDTO(HoursDTO):
    """DTO-object to transfer between hourselection and top_up"""
    hour:int
    today_hours:HourObject
    tomorrow_hours:HourObject
    prices: list
    prices_tomorrow:list


import logging

from ...util import nametoid
from ...models.const import DOMAIN
from .hubmember import HubMember
from .const import (TOTALPOWER, HOUSEPOWER)

_LOGGER = logging.getLogger(__name__)


class Power:
    def __init__(self, configsensor: str, powersensor_includes_car: bool = False):
        self._config_sensor = configsensor
        self._total = HubMember(data_type=int, initval=0, name=TOTALPOWER)
        self._house = HubMember(data_type=int, initval=0, name=HOUSEPOWER)
        self._powersensor_includes_car = powersensor_includes_car
        self._setup()

    @property
    def is_initialized(self) -> bool:
        return self._total.is_initialized and self._house.is_initialized

    @property
    def config_sensor(self) -> str:
        return self._config_sensor

    @property
    def total(self) -> HubMember:
        return self._total

    @total.setter
    def total(self, val):
        self._total = val

    @property
    def house(self) -> HubMember:
        return self._house

    @house.setter
    def house(self, val):
        self._house = val

    def _setup(self):
        if self._powersensor_includes_car is True:
            self.total.entity = self.config_sensor
            self.house.entity = nametoid(f"sensor.{DOMAIN}_{HOUSEPOWER}")
        else:
            self.house.entity = self._config_sensor

    def update(self, carpowersensor_value=0, config_sensor_value=None):
        if self._powersensor_includes_car is True:
            if config_sensor_value is not None:
                self.total.value = config_sensor_value
            self.house.value = (float(self.total.value) - float(carpowersensor_value))
        else:
            if config_sensor_value is not None:
                self.house.value = config_sensor_value
            self.total.value = (float(self.house.value) + float(carpowersensor_value))
